

#include <GL/glew.h>

#include "scene.hpp"
#include "../../lib/opengl/glutils.hpp"
#include "../../lib/perlin/perlin.hpp"
#include "../../lib/interface/camera_matrices.hpp"
#include "../interface/myWidgetGL.hpp"
#include <cmath>

#include <string>
#include <sstream>
#include "../../lib/mesh/mesh_io.hpp"

using namespace cpe;



void scene::load_scene()
{
    load_common_data();
    start_webcam();

    //get webcam dimension
    int width=0;
    int height=0;
    get_webcam_size(width,height);
    float const ratio = static_cast<float>(height)/width;


    //*****************************************//
    // Cube with webcam texture                //
    //*****************************************//

    mesh_cube = load_mesh_file("data/cube.obj");
    mesh_cube.transform_apply_translation({-0.5f, 0.0f,-0.5f});
    mesh_cube.transform_apply_scale(1.0f,ratio,1.0f);

    mesh_cube.fill_empty_field_by_default();
    mesh_cube.fill_color(vec3(1.0,1.0,1.0));

    mesh_cube_opengl.fill_vbo(mesh_cube);


    //*****************************************//
    // Ground                                  //
    //*****************************************//

    mesh_ground.add_vertex(vec3(-2.0f, 0.0f,-2.0f));
    mesh_ground.add_vertex(vec3(-2.0f, 0.0f, 2.0f));
    mesh_ground.add_vertex(vec3( 2.0f, 0.0f, 2.0f));
    mesh_ground.add_vertex(vec3( 2.0f, 0.0f,-2.0f));

    mesh_ground.add_texture_coord({0.0f,0.0f});
    mesh_ground.add_texture_coord({0.0f,1.0f});
    mesh_ground.add_texture_coord({1.0f,1.0f});
    mesh_ground.add_texture_coord({1.0f,0.0f});


    mesh_ground.add_triangle_index({0,2,1});
    mesh_ground.add_triangle_index({0,3,2});

    mesh_ground.fill_empty_field_by_default();
    mesh_ground.fill_color(vec3(0.7,0.9,0.6));

    mesh_ground_opengl.fill_vbo(mesh_ground);



}

void scene::draw_scene()
{
    prepare_shader(shader_mesh);
    glBindTexture(GL_TEXTURE_2D,texture_default);  PRINT_OPENGL_ERROR();
    mesh_ground_opengl.draw();


    prepare_shader(shader_cube);
    glUniform1f(get_uni_loc(shader_cube,"time"),time_advance);


    // Get webcam image
    cv::Mat frame;
    capture >> frame;

    // Generate OpenGL texture from webcam image
    GLuint texture_webcam = generate_texture_webcam(frame);

    // Image analysis function
    analyse_image(frame);

    // Draw the object with the webcam texture
    glBindTexture(GL_TEXTURE_2D,texture_webcam);  PRINT_OPENGL_ERROR();
    mesh_cube_opengl.draw();

    // Delete the texture on the GPU
    glDeleteTextures(1, &texture_webcam);  PRINT_OPENGL_ERROR();


}


void scene::start_webcam()
{
    capture=cv::VideoCapture(0);
    if(!capture.isOpened())
    {
        std::cerr<<"Failed to open Camera"<<std::endl;
        exit(1);
    }
}

void scene::load_common_data()
{
    texture_default=load_texture_file("data/white.jpg");

    shader_mesh = read_shader("shaders/shader_mesh.vert",
                              "shaders/shader_mesh.frag"); PRINT_OPENGL_ERROR();

    shader_cube = read_shader("shaders/shader_cube.vert",
                              "shaders/shader_cube.frag"); PRINT_OPENGL_ERROR();
}

bool pixel_is_red(float const r,float const g,float const b)
{
    if( r>1.2f*g && r>1.4f*b )
        return true;
    else
        return false;
}

void scene::analyse_image(cv::Mat const& im)
{
    int const Nu = im.rows;
    int const Nv = im.cols;

    int red_count = 0;

    int const subsample = 5;
    for(int ku=0 ; ku<Nu ; ku+=subsample)
    {
        for(int kv=0; kv<Nv ; kv+=subsample)
        {
            cv::Vec3b c = im.at<cv::Vec3b>(ku,kv);
            float const b = c.val[0];
            float const g = c.val[1];
            float const r = c.val[2];

            if( pixel_is_red(r,g,b) )
                red_count++;
        }
    }

    int const total_pixels = Nu*Nv/(subsample*subsample);
    if(static_cast<float>(red_count)/total_pixels > 0.2f)
        time_advance++;
}

void scene::prepare_shader(GLuint const shader_id)
{
    //Setup uniform parameters
    glUseProgram(shader_id);                                                                           PRINT_OPENGL_ERROR();

    //Get cameras parameters (modelview,projection,normal).
    camera_matrices const& cam=pwidget->camera();

    //Set Uniform data to GPU
    glUniformMatrix4fv(get_uni_loc(shader_id,"camera_modelview"),1,false,cam.modelview.pointer());     PRINT_OPENGL_ERROR();
    glUniformMatrix4fv(get_uni_loc(shader_id,"camera_projection"),1,false,cam.projection.pointer());   PRINT_OPENGL_ERROR();
    glUniformMatrix4fv(get_uni_loc(shader_id,"normal_matrix"),1,false,cam.normal.pointer());           PRINT_OPENGL_ERROR();

}

void scene::get_webcam_size(int &width,int &height)
{
    cv::Mat frame;
    capture>>frame;

    width = frame.cols;
    height = frame.rows;
}

GLuint scene::generate_texture_webcam(cv::Mat const& im)
{
    GLuint texture_id;
    glGenTextures(1, &texture_id);
    glBindTexture(GL_TEXTURE_2D, texture_id);  PRINT_OPENGL_ERROR();

    GLenum in_color = GL_BGR;
    if (im.channels() == 1)
        in_color = GL_LUMINANCE;

    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT); PRINT_OPENGL_ERROR();
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT); PRINT_OPENGL_ERROR();
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR); PRINT_OPENGL_ERROR();
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR); PRINT_OPENGL_ERROR();

    glTexImage2D(GL_TEXTURE_2D,0,GL_RGB,im.cols,im.rows,0,in_color,GL_UNSIGNED_BYTE,im.ptr()); PRINT_OPENGL_ERROR();

    return texture_id;
}

scene::scene()
    :shader_mesh(0),shader_cube(0),time_advance(0)
{}


GLuint scene::load_texture_file(std::string const& filename)
{
    return pwidget->load_texture_file(filename);
}

void scene::set_widget(myWidgetGL* widget_param)
{
    pwidget=widget_param;
}


