
#include <GL/glew.h>

#include "scene.hpp"
#include "../../lib/common/error_handling.hpp"
#include "../../lib/opengl/glutils.hpp"
#include "../../lib/perlin/perlin.hpp"
#include "../../lib/interface/camera_matrices.hpp"
#include "../interface/myWidgetGL.hpp"
#include <cmath>

#include <string>
#include <sstream>
#include "../../lib/mesh/mesh_io.hpp"

using namespace cpe;

void scene::load_scene()
{
    load_common_data();
    start_webcam();

    //*****************************************//
    // Background                              //
    //*****************************************//

    mesh_camera.add_vertex( { -1.0f , -1.0f , 0.5f } );
    mesh_camera.add_vertex( { -1.0f , +1.0f , 0.5f } );
    mesh_camera.add_vertex( { +1.0f , +1.0f , 0.5f } );
    mesh_camera.add_vertex( { +1.0f , -1.0f , 0.5f } );

    mesh_camera.add_texture_coord( { 1.0f , 1.0f } );
    mesh_camera.add_texture_coord( { 1.0f , 0.0f } );
    mesh_camera.add_texture_coord( { 0.0f , 0.0f } );
    mesh_camera.add_texture_coord( { 0.0f , 1.0f } );

    mesh_camera.add_triangle_index( { 0,2,1 } );
    mesh_camera.add_triangle_index( { 0,3,2 } );

    mesh_camera.fill_empty_field_by_default();
    mesh_camera_opengl.fill_vbo(mesh_camera);


    //*****************************************//
    // Cube with webcam texture                //
    //*****************************************//

    mesh_cube = load_mesh_file("data/cube.obj");
    mesh_cube.transform_apply_translation({-0.5f, -0.5f, -0.5f});
    mesh_cube.transform_apply_scale(0.2f,0.2f,0.2f);

    mesh_cube.fill_empty_field_by_default();
    mesh_cube.fill_color(vec3(1.0,1.0,1.0));

    mesh_cube_opengl.fill_vbo(mesh_cube);

}

void scene::draw_scene()
{
    // Get webcam image
    cv::Mat frame;
    capture >> frame;
    analyse_image(frame);
    GLuint texture_webcam = generate_texture_webcam(frame);

    // draw webcam image as a background
    glUseProgram(shader_webcam);                               PRINT_OPENGL_ERROR();
    glBindTexture(GL_TEXTURE_2D,texture_webcam);               PRINT_OPENGL_ERROR();
    glDepthMask(GL_FALSE);                                     PRINT_OPENGL_ERROR();
    mesh_camera_opengl.draw();
    glDepthMask(GL_TRUE);                                      PRINT_OPENGL_ERROR();

    if(is_red_active)
    {
        // draw cube if there is a red surface
        prepare_shader(shader_cube);
        mat4 rotation;
        rotation.set_rotation({0.0f,1.0f,0.0f},time_advance/20.0f);

        glUniformMatrix4fv(get_uni_loc(shader_cube,"rotation"),1,false,rotation.pointer());  PRINT_OPENGL_ERROR();
        glUniform2f(get_uni_loc(shader_cube,"translation"),middle.x(),middle.y());           PRINT_OPENGL_ERROR();
        glUniform1f(get_uni_loc(shader_cube,"scaling"),size_cube);                           PRINT_OPENGL_ERROR();
        glBindTexture(GL_TEXTURE_2D,texture_cpe);                                            PRINT_OPENGL_ERROR();

        mesh_cube_opengl.draw();
        time_advance++;
    }

    // Delete the texture on the GPU
    glDeleteTextures(1, &texture_webcam);                                                    PRINT_OPENGL_ERROR();
}



void scene::start_webcam()
{
    capture=cv::VideoCapture(0);
    if(!capture.isOpened())
    {
        std::cerr<<"Failed to open Camera"<<std::endl;
        exit(1);
    }
}

void scene::load_common_data()
{
    texture_default=load_texture_file("data/white.jpg");
    texture_cpe=load_texture_file("data/logo.jpg");

    shader_mesh = read_shader("shaders/shader_mesh.vert",
                              "shaders/shader_mesh.frag"); PRINT_OPENGL_ERROR();

    shader_cube = read_shader("shaders/shader_cube.vert",
                              "shaders/shader_cube.frag"); PRINT_OPENGL_ERROR();

    shader_webcam = read_shader("shaders/shader_webcam.vert",
                              "shaders/shader_webcam.frag"); PRINT_OPENGL_ERROR();
}

bool pixel_is_red(float const r,float const g,float const b)
{
    if( r>1.6f*g && r>1.6f*b )
        return true;
    else
        return false;
}

void scene::analyse_image(cv::Mat & im)
{
    int const Nu = im.rows;
    int const Nv = im.cols;

    int red_count = 0;
    int sum_x = 0;
    int sum_y = 0;

    int const subsample = 5;
    for(int ku=0 ; ku<Nu ; ku+=subsample)
    {
        for(int kv=0; kv<Nv ; kv+=subsample)
        {
            cv::Vec3b c = im.at<cv::Vec3b>(ku,kv);
            float const b = c.val[0];
            float const g = c.val[1];
            float const r = c.val[2];

            if( pixel_is_red(r,g,b) )
            {
                red_count++; //increment counter
                sum_x += ku; //store x coordinate
                sum_y += kv; //store y coordinate
            }
        }
    }

    int const total_pixels = Nu*Nv/(subsample*subsample);

    if(static_cast<float>(red_count)/total_pixels > 0.05f) //percent of red pixels
    {
        is_red_active=true;

        //middle position in the image
        vec2 const middle_pixel = { float(sum_x)/red_count , float(sum_y)/red_count };

        //middle position in OpenGL Texture
        vec2 const middle_screen = {1.0f-2.0f*middle_pixel.y()/Nv,
                                    1.0f-2.0f*middle_pixel.x()/Nu};

        store_middle.push_back(middle_screen);
    }
    else
        is_red_active=false;

    // The size of the cube depends on the number of red pixels
    size_cube = std::sqrt(4.0f*float(red_count)/total_pixels);

    // Manage the list storing all the previous middle position in the last frames
    if(store_middle.size()>5)
        store_middle.pop_front();

    // Compute the average middle position with respect to the last frames
    middle=vec2();
    for(vec2 const& p : store_middle)
        middle += p;
    middle /= store_middle.size();
}

void scene::prepare_shader(GLuint const shader_id)
{
    //Setup uniform parameters
    glUseProgram(shader_id);                                                                           PRINT_OPENGL_ERROR();

    //Get cameras parameters (modelview,projection,normal).
    camera_matrices const& cam=pwidget->camera();

    //Set Uniform data to GPU
    glUniformMatrix4fv(get_uni_loc(shader_id,"camera_modelview"),1,false,cam.modelview.pointer());     PRINT_OPENGL_ERROR();
    glUniformMatrix4fv(get_uni_loc(shader_id,"camera_projection"),1,false,cam.projection.pointer());   PRINT_OPENGL_ERROR();
    glUniformMatrix4fv(get_uni_loc(shader_id,"normal_matrix"),1,false,cam.normal.pointer());           PRINT_OPENGL_ERROR();
}



GLuint scene::generate_texture_webcam(cv::Mat const& im)
{
    GLuint texture_id;
    glGenTextures(1, &texture_id);
    glBindTexture(GL_TEXTURE_2D, texture_id);  PRINT_OPENGL_ERROR();

    GLenum in_color = GL_BGR;
    if (im.channels() == 1)
        in_color = GL_LUMINANCE;

    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT); PRINT_OPENGL_ERROR();
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT); PRINT_OPENGL_ERROR();
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR); PRINT_OPENGL_ERROR();
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR); PRINT_OPENGL_ERROR();

    glTexImage2D(GL_TEXTURE_2D,0,GL_RGB,im.cols,im.rows,0,in_color,GL_UNSIGNED_BYTE,im.ptr()); PRINT_OPENGL_ERROR();

    return texture_id;
}

scene::scene()
    :shader_mesh(0),shader_cube(0),time_advance(0)
{}


GLuint scene::load_texture_file(std::string const& filename)
{
    return pwidget->load_texture_file(filename);
}

void scene::set_widget(myWidgetGL* widget_param)
{
    pwidget=widget_param;
}


